/* klunar.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   Linux kernel module wrapper for the LUNAR logic
 *
 * Copyright (C) 2001-2004 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *   2004-01-09 moved generic init stuff into src/lunar_init.c,
 *              added mackill parameter parsing
 *   2004-01-05 the sources are now managed by a CVS
 *   2003-11-20 adjusted debuglevels of printk messages, init/fin of
 *              module now gets announced
 */

#define __KERNEL__
#define __OPTIMIZE__
#define MODULE

#include "../include/platform.h"
#include "../include/version.h"

#include <net/neighbour.h>
#ifndef CONFIG_ARM
# include <net/arp.h>
#else
extern struct neigh_table arp_tbl;
# include <endian.h>
#endif

#include "../include/sapf.h"
#include "../include/xrp.h"
#include "../include/netbox.h"
#include "../include/lunar.h"
#include "../include/sapf_decl.h"
#include "../include/xrp_decl.h"
#include "../include/netbox_decl.h"
#include "../include/lunar_decl.h"

int conf_min_ring = LUNAR_MIN_RING;
int conf_max_ring = LUNAR_MAX_RING;
int conf_timeout_per_ring = LUNAR_TIMEOUT_PER_RING;
int conf_max_retries = LUNAR_MAX_RETRIES;
static int conf_fri = LUNAR_RRET;           // forced rediscovery interval
int conf_fri_jitter = LUNAR_RRET_JITTER;
int conf_rreq_gap = 0;
int conf_data_gap = 0;

int xrp_param_len[XRP_PARTYPE_COUNT] =
{       // XRP_PARTYPE_UNKNOWN:
	0,
	// XRP_PARTYPE_SEL:
	sizeof(struct xrp_head_s) + sizeof(sapfsel_t),
	// XRP_PARTYPE_IPv4
	sizeof(struct xrp_head_s) + sizeof(uint32_t),
	// XRP_PARTYPE_IPv6
	sizeof(struct xrp_head_s) + 16,
	// XRP_PARTYPE_SEL_ETH
	sizeof(struct xrp_head_s) + sizeof(sapfsel_t) + ETH_ALEN,
	// XRP_PARTYPE_SEL_UDP
	sizeof(struct xrp_head_s) + sizeof(sapfsel_t) + sizeof(uint32_t)
						      + sizeof(short),
	// XRP_PARTYPE_HID128
	sizeof(struct xrp_head_s) + 16
};

#define MAX_MACKILL 3
static int mackill_cnt;
static unsigned char mackill_list[MAX_MACKILL][ETH_ALEN];

struct peer_s peers[LUNAR_MAX_PEERS];

#include "../src/lunar_init.c"
#include "knetbox.c"

// ----------------------------------------------------------------------

MODULE_AUTHOR("christian.tschudin@unibas.ch");
MODULE_DESCRIPTION("LUNAR kernel module (" LUNAR_VERSION "), 2002-2004");
MODULE_LICENSE("GPL");

MODULE_PARM(basedevname, "s");
MODULE_PARM_DESC(basedevname, "name of base device (default: first device "
		          "with wireless extensions)");

MODULE_PARM(debuglevel, "i");
MODULE_PARM_DESC(debuglevel, "debug level 0-10");

MODULE_PARM(forwarding, "s");
MODULE_PARM_DESC(forwarding, "default gatewaying (no, yes)");

MODULE_PARM(mackill, "s");
MODULE_PARM_DESC(mackill, "suppress packets from the given MAC address/es");
static char *mackill;

// ----------------------------------------------------------------------

static int klunar_netdev_event(struct notifier_block *this,
			       unsigned long event, void *ptr);
static struct notifier_block klunar_notifier = {
	notifier_call: klunar_netdev_event,
};

static int klunar_netdev_event(struct notifier_block *this,
                              unsigned long event, void *ptr)
{
        struct net_device *dev = (struct net_device *) ptr;

	if (!dev)
		return NOTIFY_DONE;

        DEBUG(10, printk("klunar notifier event %08lx %s\n", event, dev->name));

        switch (event) {
        case NETDEV_REGISTER:
        case NETDEV_UP:
		DEBUG(10, printk("notifier up %s\n", dev->name));
		if (basedev)
			break;
		if (basedevname) {
			struct net_device *d = dev_get_by_name(basedevname);
			if (!d)
				break;
			dev_put(d);
			if (d->ifindex != dev->ifindex)
				break;
			if (!strcmp(basedevname, dev->name)) {
				DEBUG(1, printk("klunar - new interface %s cannot be the lunar device\n", basedevname));
				break;
			}
		} else {
			if (!dev->get_wireless_stats)
				break;
		}
		DEBUG(1, printk("klunar - new (wireless) interface %s found\n",
				 dev->name));
		basedev = dev;
		netbox_addL2if(dev->ifindex, myhandle);
                break;
        case NETDEV_UNREGISTER:
        case NETDEV_DOWN:
		DEBUG(10, printk("notifier down %s\n", dev->name));
                if (dev == basedev) {
                        DEBUG(1, printk("klunar - (wireless) interface %s went away\n",
					 dev->name));
			basedev = 0;
			netbox_dropL2if(dev->ifindex);
                }
                break;

        default:
                break;
        };

        return NOTIFY_DONE;
}

// ----------------------------------------------------------------------

#include "../src/lunar.c"
#include "../src/lunar_dhcp.c"
#include "../src/xrp.c"

// ----------------------------------------------------------------------

static int parse_mackill()
{
	char *pa[MAX_MACKILL], *cp;
	int i, j; // , ia[ETH_ALEN];

	DEBUG(10, printk("mackill parse: %s\n", mackill));

	while (mackill_cnt < MAX_MACKILL) {
		pa[mackill_cnt] = strtok(mackill_cnt ? NULL : mackill, ",");
		if (!pa[mackill_cnt])
			break;
		mackill_cnt++;
	}
	for (i = 0; i < mackill_cnt; i++) {
		// lnx kernel bug in 2.4.X: sscanf format "%x" does not work ....
		cp = pa[i];
		for (j = 0; j < ETH_ALEN; j++, cp++) {
			mackill_list[i][j] = 0;
			for (; isxdigit(*cp); cp++) {
				mackill_list[i][j] =
					(mackill_list[i][j] << 4) |
					(*cp <= '9' ? (*cp - '0') :
					           ((*cp & 0x07) + 9 ));
			}
			if (*cp && *cp != ':') break;
		}
		if ( j != ETH_ALEN) {
			DEBUG(0, printk("mackill: error in MAC addr %s\n", pa[i]));
			mackill_cnt--;
			return -1;
		}

		DEBUG(5, printk("mackill +%s\n",
				netbox_eth2str(mackill_list[i])));
	}
	return 0;
}

// ----------------------------------------------------------------------

static int klunar_open(struct net_device *dev)
{
	netif_start_queue(dev);
        return 0;
}


static int klunar_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
        struct net_device_stats *stats = (struct net_device_stats *) dev->priv;
	sapfsel_t sel;
	sapfpkt pkt;

        DEBUG(10, printk("klunar start_xmit (%d bytes)\n", skb->len));

        switch (ntohs(*(unsigned short *) (skb->data + 12))) {
	case ETH_P_ARP:
		DEBUG(10, printk("klunar - arp detected\n"));
		lunar_do_arp(*(struct in_addr *)(skb->data + 38));
                kfree_skb(skb);
		break;
	case ETH_P_IP:
		if (stats) {
			stats->tx_packets++;
			stats->tx_bytes += skb->len;
		}
	{
		int i;
		DEBUG(10, printk("IP data (%d):", skb->len));
		for (i = 0; i < 50; i++)
			DEBUG(10, printk(" %02x", skb->data[i]));
		DEBUG(10, printk("\n"));
	}
		pkt.data = skb->data + dev->hard_header_len;
		pkt.len = skb->len - dev->hard_header_len;
		pkt.ifnr = -1;

		if (pkt.len >= (20+8+44)  && *((char*)pkt.data+9) == 17 && // UDP
		    ntohs(*(short*)((char*)pkt.data+20)) == 68 &&  // DHCP clnt
		    ntohs(*(short*)((char*)pkt.data+22)) == 67)    // DHCP srvr
			lunar_do_dhcp(&pkt);
		else if (!memcmp(skb->data, "\xff\xff\xff\xff\xff\xff", ETH_ALEN)){ // broadcast
			if (bcast_selector)
				sapf_demux(mysapfhandle, bcast_selector, &pkt);
			else
				lunar_do_bc_setup(&bcast_selector, &pkt);
		} else { // unicast
			sel = 0;
			memcpy(((char*)&sel) + sizeof(sel) - ETH_ALEN,
			       skb->data,ETH_ALEN);
			sapf_demux(mysapfhandle, sel, &pkt);
		}
		kfree_skb(skb);
		break;
	default:
		kfree_skb(skb);
		break;
	}

        return 0;
}


static int klunar_stop(struct net_device *dev)
{
	if (basedev) {
		int i = basedev->ifindex;
		basedev = 0;
		netbox_dropL2if(i);
	}
        netif_stop_queue(dev);

        return 0;
}


static struct net_device_stats *klunar_get_stats(struct net_device *dev)
{
        return (struct net_device_stats *) dev->priv;
}


// ----------------------------------------------------------------------

static int __init init_klunar(void)
{
	struct net_device *dev = 0;

	DEBUG(1, printk("klunar init module\n"));

	parse_mackill();

        lunardev = init_etherdev(0, sizeof(struct net_device_stats));
        if (!lunardev) {
		DEBUG(10, printk("klunar_init: no lunardev\n"));
		sapf_close(mysapfhandle);
                return -ENOMEM;
	}
	SET_MODULE_OWNER(lunardev);
	DEBUG(1, printk("klunar - lunar device is %s\n", lunardev->name));

	if (basedevname) {
		dev = dev_get_by_name(basedevname);
		if (!dev) {
			DEBUG(1, printk("klunar - device %s not found\n", basedevname));
		} else {
			dev_put(dev);
			if (dev == lunardev) {
				DEBUG(1, printk("klunar - invalid base device %s\n",
				       basedevname));
				dev = 0;
			}
		}
		basedev = dev;
		if (dev)
			netbox_addL2if(dev->ifindex, mysapfhandle);
	} else {
		read_lock(&dev_base_lock);
		for (dev = dev_base; dev != NULL; dev = dev->next) {
			if (dev->get_wireless_stats)
				break;
		}
		if (dev) {
			basedev = dev;
			DEBUG(10, printk("klunar - wireless interface is %s\n",
				 dev->name));
			netbox_addL2if(dev->ifindex, mysapfhandle);
		}
		read_unlock(&dev_base_lock);
	}

        lunardev->open = &klunar_open;
        lunardev->hard_start_xmit = &klunar_start_xmit;
        lunardev->stop = &klunar_stop;
        lunardev->get_stats = &klunar_get_stats;

	lunardev->mtu -= sizeof(sapfsel_t);
        lunardev->tx_queue_len = 3;
        *(short*)(lunardev->dev_addr) = htons(0x00ff);
        get_random_bytes(lunardev->dev_addr + 2, 4);

	if (lunar_init()) {
		unregister_netdev(lunardev);
		return -1;
	}

        register_netdevice_notifier(&klunar_notifier);
        proc_net_create("lunar", 0, lunar_proc_info);

	return 0;
}


static void __exit exit_klunar(void)
{
	DEBUG(1, printk("klunar unloading module\n"));

        proc_net_remove("lunar");
        unregister_netdevice_notifier(&klunar_notifier);
	unregister_netdev(lunardev);

	lunar_exit();
}


module_init(init_klunar);
module_exit(exit_klunar);

// eof
