/* ksapf.c
 *
 *   SAPF kernel module
 *
 *   The Linux kernel module wrapper for
 *   a partial implementation of the Simple Active Packet Format,
 *
 * Copyright (C) 2001-2003 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *   2003-11-02 some bug fixes
 *   2003-05-10 added support for multiple named fib tables
 *   2002-12-24 selectors now stored in network order, removed two '&&' bugs
 *   2002-12-21 permit NULL callback ie black hole (or use a redirect to 0)
 *   2002-12-03 added activity flag, export the new testandreset symbol
 *   2002-10-11 added multi-client support (explicit open, returns handle)
 *   2002-09-17 fixed infinite loop in hash table fine, discovered by
                David Rosen and Olof Rensfelt
 *   2002-06-20 kernelized for Linux
 *   2001-11-18 initial version for SelNet
 */

#define __KERNEL__
#define __OPTIMIZE__
#define MODULE

#include "../include/platform.h"
#include "../include/version.h"
#include "../src/sapf.c"

MODULE_AUTHOR("christian.tschudin@unibas.ch");
MODULE_DESCRIPTION("SAPF kernel module, May 2003");
MODULE_LICENSE("GPL");

MODULE_PARM(debuglevel, "i");
MODULE_PARM_DESC(debuglevel, "debug level 0-10");

// ----------------------------------------------------------------------

static int sapf_proc_info(char *buf, char **start, off_t offset, int len)
{
	int written = 0, t;

#define adjustlen()	{ \
	if (written < offset) { offset -= written; written = 0; } \
	if (len <= written) \
		goto out; \
}

	for (t = 0; t < SAPF_MAX_FIB_TABLES; t++) {
		int i, cnt;
		struct fibtable_s *ftab = fib[t];
		struct fibentry_s *f;

		if (!ftab)
			continue;

		for (f = ftab->tab, i = ftab->fibsize, cnt = 0; i > 0; f++, i--)
			if ((f->flags_and_op & FE_MASK) == FE_USED)
				cnt++;

		written += sprintf(buf, "# sapf table `%s' (%d table entries, %d used)\n", ftab->name, ftab->fibsize, cnt);

		for (f = ftab->tab, i = 0; i < ftab->fibsize; i++, f++) {
			if ((f->flags_and_op & FE_MASK) == FE_USED) {
				written += sprintf(buf + written,
					   "%02d %016Lx %5d  ",
					   f->client, ntohsel(f->sel), f->credit);
				adjustlen();
				switch (f->flags_and_op & SAPFOP_MASK) {
				case SAPFOP_CBACK:
					written += sprintf(buf + written,
					     "cb %p %p\n", f->p1.callback,
					     f->p2.aux);
					break;
				case SAPFOP_REDIR:
					written += sprintf(buf + written,
					     "-> %016Lx\n", ntohsel(f->p1.sel));
					break;
				default:
					written += sprintf(buf + written, "?\n");
					break;
				}
				adjustlen();
			}
		}
	}
out:
	*start = buf + offset;
	written -= offset;
	if (written > len)
		return len;
	return (written < 0) ? 0 : written ;
}

static int __init ksapf_init(void)
{
        printk("ksapf init module\n");

	if (sapf_init())
		return -1;

        proc_net_create("sapf", 0, sapf_proc_info);

	return 0;
}

static void __exit ksapf_fini(void)
{
	printk("ksapf unloading module\n");

        proc_net_remove("sapf");

	sapf_fini();
}


EXPORT_SYMBOL(sapf_open);
EXPORT_SYMBOL(sapf_close);
EXPORT_SYMBOL(sapf_demux);
EXPORT_SYMBOL(sapf_getUnusedSelector);
EXPORT_SYMBOL(sapf_exists);
EXPORT_SYMBOL(sapf_addCallback);
EXPORT_SYMBOL(sapf_getCallbackAux);
EXPORT_SYMBOL(sapf_setPermanent);
EXPORT_SYMBOL(sapf_deleteEntry);
EXPORT_SYMBOL(sapf_addREDIR);
EXPORT_SYMBOL(sapf_testandreset_active);

module_init(ksapf_init);
module_exit(ksapf_fini);

// eof
